//  This is the JavaScript file that accompanies tictactoe.html and tictactoe.css.
//  Together these provide a web app where the user plays tic-tac-toe against the computer.


/*  This function is used for validation of the data placed into the tiny text boxes
 *  within the tic-tac-toe grid.  Only X, O, or the empty string are accepted.
 */
function check(location)
   {
   var textbox = document.getElementById(location);
   var val = textbox.value;

   switch (val)
      {
      case "X":
      case "O":
      case "":
         return true;
      default:
         {
         alert("Invalid value in that last location.\nUse X or O. Try again.");
         textbox.value = "";  // Don't allow an invalid value.  Replace with empty string.
         return false;
         }
      }
   }


/*  This function was used for debugging purposes only and is not currently used in the web app.
 *  It placed the contents of the board array (representing the state of the tic-tac-toe game)
 *  into a paragraph on the web page.
 */
function Display(board)
   {
   var msg = document.getElementById("message");
   var s = "board: ";
   for (var k = 0; k < board.length; k++)
      {
      s = s + board[k] + " ";
      }
   msg.innerHTML = s;
   }


/*  This function is the event handler for when the user clicks on the Start button.
 *  It figures out who goes first in the game and then gets the game started.
 */
function Start()
   {
   var player;
   // The number of free places in the grid is kept at index 0 of board. At index 10 is N = no result, D = draw, W = win.
   var board = new Array('9', '', '', '', '', '', '', '', '', '', 'N');
   var msg = document.getElementById("message");
   var choice = document.getElementById("choice");
   var item;

   if (document.getElementById("Xfirst").checked)
      {
      player = 'X';
      choice.style.visibility = "hidden";
      Game(player, board);   // Play the game, starting with the computer.
      }
   else if (document.getElementById("Ofirst").checked)
      {
      player = 'O';
      choice.style.visibility = "hidden";
      Game(player, board);   // Play the game, starting with the user.
      }
   else
      msg.innerHTML = "You need to select who goes first and then click on Start.";
   
   return false;
   }


/*  Calling this function causes the playing of the tic-tac-toe game on the given board
 *  with player making the first move.
 */
function Game(player, board)
   {
   var msg = document.getElementById("message");
   var MoveButton = document.getElementById("MoveButton");

   if (player == 'X')   // The computer's turn
      {
      MoveButton.style.visibility = "hidden";
      msg.innerHTML = "My turn (X).";   
      if (ExecuteBestComputerMove(board)) 
         {
         board[0]--;
         }
      else
         {
         msg.innerHTML = "Error: game has failed";
         } 
      if (! GameDone('X', board))
         Game('O', board);
      }
   else if (player == 'O')   // The user's turn
      {
      MoveButton.style.visibility = "visible";
      msg.innerHTML = "Your turn. Put O in the desired square and click the Move button.";
      // The rest of the user's turn will be processed by the Move button click handler.
      }
   }


/*  This is the click handler for the Move button, the button that the user clicks to register a move of O
 *  to a particular square on the board.  It uses the current data to decide whether the game is over and if
 *  not to continue with the game.
 */
function UserMove()
   {
   // Look up what X's and O's are in the tic-tac-toe text boxes so as to create the current board array:
   var board = new Array(11);
   var item;
   var count = 0;

   for (var k = 1; k <= 9; k++)
      {
      item = document.getElementById(k);
      if ((typeof(item.value) == "undefined") || (typeof(item.value) == "object"))
         {
         item.value = '';
         }
      else if ((item.value != 'X') && (item.value != 'O'))
         {
         item.value = '';
         count++;
         }

      board[k] = item.value;
      }

   board[0] = count;
   board[10] = 'N';  // though this may change when GameDone checks for a win or draw

   if (! GameDone('O', board))
      Game('X', board);

   return false;
   }


/*  This function decides whether or not the game is over due to player winning or
 *  to a draw.  (Player is the last side that made a move: X or O.)  This is done
 *  by examining the X's and O's on the current board as well as the number of
 *  free locations (stored in board[0]) and any known status of the game
 *  (as recorded in board[10]).  A true is returned if the game is over, false otherwise.
 *  The function also displays appropriate messages if the game is over and hides the 
 *  Move button in that case.
 */
function GameDone(player, board)
   {
   var msg = document.getElementById("message");
   var won = Won(player, board);
   var draw = Draw(player, board);

   if (won || draw)
      {
      var MoveButton = document.getElementById("MoveButton");
      MoveButton.style.visibility = "hidden";
      }

   if (won)
      {
      msg.innerHTML = player + " has won.";
      return true;
      }
   else if (draw)
      {
      msg.innerHTML = "The game is a draw.";
      return true;
      }
   else
      {
      return false;
      }
   }


/*  This function takes player (X or O, whichever made the last move) and board (the array of 
 *  X's, O's, and empty strings) and returns true or false to indicate if player has won.
 */
function Won(player, board)
   {
   // Quick way to check if X has won:
   if (player == 'X')
      {
      if (board[10] == 'W')
         return true;
      }
   else   // player is O
      {
      // To see if O has won is more laborious:
      var val1, val2, val3, num1, num2, num3;
      var slices = [[1, 2, 3], [4, 5, 6], [7, 8, 9], [1, 4, 7], [2, 5, 8], [3, 6, 9], [9, 5, 1], [7, 5, 3]];

      for (var k = 0; k < 8; k++)
         {
         num1 = slices[k][0];
         num2 = slices[k][1];
         num3 = slices[k][2];
         val1 = board[num1];
         val2 = board[num2];
         val3 = board[num3];

         if ((val1 == 'O') && (val2 == 'O') && (val3 == 'O'))
            {
            return true;
            }
         }   
      }

   return false;
   }


/*  This function takes the same parameters player and board as the previous function and returns true or false
 *  to indicate if the game is (or is not) a draw.
 */
function Draw(player, board)
   {
   if (board[0] == 0)  // This assumes we have already checked to see if someone won. If no one has won and the board is filled,
      return true;     // then there is a draw.
   else
      return false;
   }


/*  This function uses the board array to select the best move for the computer to make.
 *  By "best" this means the best choice according to the algorithms in use here.
 *  In fact, these algorithms do not follow optimal tic-tac-toe strategy.
 */
function ExecuteBestComputerMove(board)
   {
   if (WinningMove(board))
      return true;
   if (ForcedMove(board))
      return true;
   if (FirstDiagonalMove(board))
      return true;
   if (SecondDiagonalMove(board))
      return true;
   if (GoodMove(board))
      return true;
   if (CenterMove(board))
      return true;
   if (CornerMove(board))
      return true;
   if (OpenMove(board))
      return true;
   return false;
   }


/*  This function checks to see if X (the computer) can win on this move and if so, makes that move
 *  by updating the board array and the correct text box in the grid on the web page.
 */
function WinningMove(board)
   {
   var val1, val2, val3, num1, num2, num3, item;
   var slices = [[1, 2, 3], [4, 5, 6], [7, 8, 9], [1, 4, 7], [2, 5, 8], [3, 6, 9], [9, 5, 1], [7, 5, 3]];

   for (var k = 0; k < 8; k++)
      {
      num1 = slices[k][0];
      num2 = slices[k][1];
      num3 = slices[k][2];
      val1 = board[num1];
      val2 = board[num2];
      val3 = board[num3];

      if ((val1 != 'X') && (val1 != 'O') && (val2 == 'X') && (val3 == 'X'))
         {
         item = document.getElementById(num1);
         item.value = 'X';
         board[num1] = 'X';
         board[10] = 'W';
         return true;
         }
      else if ((val2 != 'X') && (val2 != 'O') && (val1 == 'X') && (val3 == 'X'))
         {
         item = document.getElementById(num2);
         item.value = 'X';
         board[num2] = 'X';
         board[10] = 'W';
         return true;
         }
      else if ((val3 != 'X') && (val3 != 'O') && (val1 == 'X') && (val2 == 'X'))
         {
         item = document.getElementById(num3);
         item.value = 'X';
         board[num3] = 'X';
         board[10] = 'W';
         return true;
         }
      }

   return false;
   }


/*  This function checks the state of the board array to see if X (the computer) must block a possible win by O.
 *  If so, that blocking move is carried out by updating the board array and the correct text box in the grid on the web page.
 */
function ForcedMove(board)
   {
   var val1, val2, val3, num1, num2, num3, item;
   var slices = [[1, 2, 3], [4, 5, 6], [7, 8, 9], [1, 4, 7], [2, 5, 8], [3, 6, 9], [9, 5, 1], [7, 5, 3]];
   
   for (var k = 0; k < 8; k++)
      {
      num1 = slices[k][0];
      num2 = slices[k][1];
      num3 = slices[k][2];
      val1 = board[num1];
      val2 = board[num2];
      val3 = board[num3];

      if ((val1 != 'X') && (val1 != 'O') && (val2 == 'O') && (val3 == 'O'))
         {
         item = document.getElementById(num1);
         item.value = 'X';
         board[num1] = 'X';
         return true;
         }
      else if ((val2 != 'X') && (val2 != 'O') && (val1 == 'O') && (val3 == 'O'))
         {
         item = document.getElementById(num2);
         item.value = 'X';
         board[num2] = 'X';
         return true;
         }
      else if ((val3 != 'X') && (val3 != 'O') && (val1 == 'O') && (val2 == 'O'))
         {
         item = document.getElementById(num3);
         item.value = 'X';
         board[num3] = 'X';
         return true;
         }
      }

   return false;
   }


/*  Given the position number for one of the corners of the board array, this function
 *  finds and returns an array of the two neighboring positions on the board.
 */
function neighbors(position)
   {
   if (position == 1)
      return [2, 4];
   if (position == 3)
      return [2, 6];
   if (position == 7) 
      return [4, 8];
   if (position == 9)
      return [6, 8];

   return [];   //*** If this function is used correctly, this line will never be reached.
   }


/*  Given the board array, this function checks to see if it is possible for X to move to a corner where
 *  both neighbors hold O's and the diagonal ending at the corner has one X already in it, with the other
 *  cell open.  If so, this move is made.
 */
function FirstDiagonalMove(board)
   {
   var val1, val2, val3, num1, num2, num3, item;
   var diagonals = [[1, 5, 9], [9, 5, 1], [3, 5, 7], [7, 5, 3]];

   for (var k = 0; k < 4; k++)
      {
      num1 = diagonals[k][0];
      num2 = diagonals[k][1];
      num3 = diagonals[k][2];
      val1 = board[num1];
      val2 = board[num2];
      val3 = board[num3];
      var nbrs = neighbors(num3);
      var valA = board[nbrs[0]];
      var valB = board[nbrs[1]];

      if ((val2 != 'X') && (val2 != 'O') && (val1 == 'X') && (val3 != 'X') && (val3 != 'O') && (valA == 'O') && (valB == 'O'))
         {
         item = document.getElementById(num3);
         item.value = 'X';
         board[num3] = 'X';
         return true;
         }
      else if ((val1 != 'X') && (val1 != 'O') && (val2 == 'X') && (val3 != 'X') && (val3 != 'O') && (valA == 'O') && (valB == 'O'))
         {
         item = document.getElementById(num3);
         item.value = 'X';
         board[num3] = 'X';
         return true;
         }
      }

   return false;
   }


/*  Given the board array, this function checks to see if it is possible for X to move to a corner where
 *  one neighbor holds a O, the other neighbor is empty, and the diagonal ending at the corner has one X
 *  already in it, with the other cell open.  If so, this move is made.
 */
function SecondDiagonalMove(board)
   {
   var val1, val2, val3, num1, num2, num3, item;
   var diagonals = [[1, 5, 9], [9, 5, 1], [3, 5, 7], [7, 5, 3]];

   for (var k = 0; k < 4; k++)
      {
      num1 = diagonals[k][0];
      num2 = diagonals[k][1];
      num3 = diagonals[k][2];
      val1 = board[num1];
      val2 = board[num2];
      val3 = board[num3];
      var nbrs = neighbors(num3);
      var valA = board[nbrs[0]];
      var valB = board[nbrs[1]];

      if ((val2 != 'X') && (val2 != 'O') && (val1 == 'X') && (val3 != 'X') && (val3 != 'O') && (valA == 'O')
         && (valB != 'X') && (valB != 'O'))
         {
         item = document.getElementById(num3);
         item.value = 'X';
         board[num3] = 'X';
         return true;
         }
      else if ((val1 != 'X') && (val1 != 'O') && (val2 == 'X') && (val3 != 'X') && (val3 != 'O') && (valB == 'O') 
         && (valA != 'X') && (valA != 'O'))
         {
         item = document.getElementById(num3);
         item.value = 'X';
         board[num3] = 'X';
         return true;
         }
      else if ((val2 != 'X') && (val2 != 'O') && (val1 == 'X') && (val3 != 'X') && (val3 != 'O') && (valB == 'O') 
         && (valA != 'X') && (valA != 'O'))
         {
         item = document.getElementById(num3);
         item.value = 'X';
         board[num3] = 'X';
         return true;
         }
      else if ((val1 != 'X') && (val1 != 'O') && (val2 == 'X') && (val3 != 'X') && (val3 != 'O') && (valA == 'O') 
         && (valB != 'X') && (valB != 'O'))
         {
         item = document.getElementById(num3);
         item.value = 'X';
         board[num3] = 'X';
         return true;
         }
      }

   return false;
   }


/*  Given the state of the board array, this function checks to see if the computer can get 2 X's in one row
 *  with the remaining spot in that row open.  If so, this move is made.
 */
function GoodMove(board)
   {
   var val1, val2, val3, num1, num2, num3, item;
   var slices = [[1, 2, 3], [4, 5, 6], [7, 8, 9], [1, 4, 7], [2, 5, 8], [3, 6, 9], [9, 5, 1], [7, 5, 3]];

   for (var k = 0; k < 8; k++)
      {
      num1 = slices[k][0];
      num2 = slices[k][1];
      num3 = slices[k][2];
      val1 = board[num1];
      val2 = board[num2];
      val3 = board[num3];

      if ((val2 == 'X') && (val1 != 'X') && (val1 != 'O') && (val3 != 'O') && (val3 != 'O'))
         {
         item = document.getElementById(num1);
         item.value = 'X';
         board[num1] = 'X';
         return true;
         }
      else if ((val1 == 'X') && (val2 != 'X') && (val2 != 'O') && (val3 != 'X') && (val3 != 'O'))
         {
         item = document.getElementById(num3);
         item.value = 'X';
         board[num3] = 'X';
         return true;
         }
      else if ((val3 == 'X') && (val1 != 'X') && (val1 != 'O') && (val2 != 'O') && (val2 != 'O'))
         {
         item = document.getElementById(num1);
         item.value = 'X';
         board[num1] = 'X';
         return true;
         }
      }

   return false;
   }


/*  This function checks the state of the board array to see if the center square is open.
 *  If so, X makes this move to the center square.
 */
function CenterMove(board)
   {
   if ((board[5] != 'X') && (board[5] != 'O'))
      {
      var item = document.getElementById(5);
      item.value = 'X';
      board[5] = 'X';
      return true;
      }

   return false;
   }


/*  Given the state of the board array, this function decides if there is an open corner.
 *  If so, X makes the move to the first open corner found.
 */
function CornerMove(board)
   {
   var item;

   if ((board[1] != 'X') && (board[1] != 'O'))
      {
      item = document.getElementById(1);
      item.value = 'X';
      board[1] = 'X';
      return true;
      }

   if ((board[3] != 'X') && (board[3] != 'O'))
      {
      item = document.getElementById(3);
      item.value = 'X';
      board[3] = 'X';
      return true;
      }

   if ((board[7] != 'X') && (board[7] != 'O'))
      {
      item = document.getElementById(7);
      item.value = 'X';
      board[7] = 'X';
      return true;
      }

   if ((board[9] != 'X') && (board[9] != 'O'))
      {
      item = document.getElementById(9);
      item.value = 'X';
      board[9] = 'X';
      return true;
      }

   return false;
   }


/*  This function, given the state of the board array, decides whether there is an open
 *  square.  If so, X makes the move to the first open square found.
 */
function OpenMove(board)
   {
   var item;

   if ((board[2] != 'X') && (board[2] != 'O'))
      {
      item = document.getElementById(2);
      item.value = 'X';
      board[2] = 'X';
      return true;
      }

   if ((board[4] != 'X') && (board[4] != 'O'))
      {
      item = document.getElementById(4);
      item.value = 'X';
      board[4] = 'X';
      return true;
      }

   if ((board[6] != 'X') && (board[6] != 'O'))
      {
      item = document.getElementById(6);
      item.value = 'X';
      board[6] = 'X';
      return true;
      }

   if ((board[8] != 'X') && (board[8] != 'O'))
      {
      item = document.getElementById(8);
      item.value = 'X';
      board[8] = 'X';
      return true;
      }

   return false;
   }


/*  This function is the click handler for the Restart button on the web page.
 *  It wipes out any message in the paragraph at the bottom, makes the move button
 *  invisible, and makes visible the opening choice of who goes first along with the
 *  Start button to begin the game.
 */
function Restart()
   {
   document.getElementById("choice").style.visibility = "visible";
   document.getElementById("message").innerHTML = "";
   document.getElementById("MoveButton").style.visibility = "hidden";
   }

